import numpy as np
import torch
import torch.nn as nn

import os

import matplotlib.pyplot as plt
plt.ion()

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

def analysis(name,mult,dim,N=2**14,epochs=2**12,lr0=-1,lr1=-2, plot=True, training_σ=1):
	mult = mult.to(device)
	lrs = 10**(torch.linspace(lr0,lr1,epochs))
	optim = torch.optim.Adam(mult.parameters(),lr=1e-2,eps=1e-12)
	hist = []
	for epoch in range(epochs):
		mult.train()
		lr = (lrs[epoch]).item()
		for g in optim.param_groups: g['lr'] = lr
		### many matrices
		A = torch.randn(N,dim,dim,device=device)*training_σ # 10
		B = torch.randn(N,dim,dim,device=device)*training_σ # 10
		### matrix multiplication
		y = A@B # = torch.einsum('nij,njk->nik',A,B) # ground truth
		Y = mult(A,B)
		### training
		loss = nn.MSELoss()(y,Y)
		optim.zero_grad()
		loss.backward()
		optim.step()
		hist.append(loss.item())
	if plot: plt.scatter(range(len(hist)),np.log(hist),s=.05,label=name)
	mult.hist = hist
	return mult

def viz_error(mult,dim,N=2**14):
		A = torch.randn(N,dim,dim,device=device)
		B = torch.randn(N,dim,dim,device=device)
		y = A@B # = torch.einsum('nij,njk->nik',A,B) # ground truth
		Y = mult(A,B)
		plt.hist(torch.flatten(torch.log((y-Y)**2)).detach().cpu(),128,density=True)
		# plt.hist(torch.flatten(((y-Y)**2)).detach().cpu(),128,density=True)

###############################################################################
### MODELS ####################################################################
###############################################################################

import anyact as aa

### this is obviously crap
# class MatMul_0(nn.Module):
# 	def __init__(self,dim):
# 		super().__init__()
# 		self.dim=dim
# 		self.flatten = nn.Flatten()
# 		self.model = nn.Linear(2*dim*dim,dim*dim)
# 	def forward(self,A,B):
# 		AB = torch.hstack((self.flatten(A),self.flatten(B)))
# 		return self.model(AB).view(-1,self.dim,self.dim)

### this is slightly less crap
class MatMul_0(nn.Module):
	def __init__(self,dim_left,dim_right,depth=5,width=1):
		super().__init__()
		if not dim_left[1]==dim_right[0]: print("ERROR INCOMPATIBILITY DETECTED")
		self.dim_left = dim_left
		self.dim_right = dim_right
		self.flatten = nn.Flatten()
		dim_in  = dim_left[0]*dim_left[1] + dim_right[0]*dim_right[1]
		dim_mid = width*dim_in
		dim_out = dim_left[0]*dim_right[1]
		self.model = nn.Sequential(*([
			# ] + (depth>0)*[ nn.Linear(dim_in ,dim_mid)  , aa.Stack([nn.Identity(), nn.Sequential(nn.GELU(),nn.BatchNorm1d(dim_mid))]), # nn.BatchNorm1d(2*dim_mid),
			# ] + (depth-1)*[ nn.Linear(2*dim_mid,dim_mid), aa.Stack([nn.Identity(), nn.Sequential(nn.GELU(),nn.BatchNorm1d(dim_mid))]), # nn.BatchNorm1d(2*dim_mid),
			] + (depth>0)*[ nn.Linear(dim_in ,dim_mid)  , aa.Stack([nn.Identity(), nn.GELU()]), # nn.BatchNorm1d(2*dim_mid),
			] + (depth-1)*[ nn.Linear(2*dim_mid,dim_mid), aa.Stack([nn.Identity(), nn.GELU()]), # nn.BatchNorm1d(2*dim_mid),
			] + (depth>0)*[ nn.Linear(2*dim_mid,dim_out),
			] + (depth<1)*[ nn.Linear(dim_in ,dim_out),
			]
		))
	def forward(self,A,B):
		AB = torch.hstack((self.flatten(A),self.flatten(B)))
		return self.model(AB).view(-1,self.dim_left[0],self.dim_right[1])

# A = torch.randn(5,10,20)
# B = torch.randn(5,20,5)
# L = MatMul_0b(A.shape[1:],B.shape[1:],depth=5,width=1)
# L(A,B)

plt.ioff()

os.makedirs(f"out/",exist_ok=True)
plt.clf()
dim=1
for depth in [1,2,3,4,5,10]:
	width=8
	mult = MatMul_0([dim,dim],[dim,dim],depth=depth,width=width)
	analysis(f"depth={depth}, width={width}",mult,dim,N=2**14,epochs=2**12,training_σ=10)

plt.xscale('log')
plt.ylabel("log(MSE)")
plt.xlabel("epoch")
plt.title("scalar multiplication by depth")
plt.legend()
plt.savefig("out/matmul_scalar_by_depth.png")


plt.clf()
dim=1
for width in [1,2,3,4,5,10]:
	depth=5
	mult = MatMul_0([dim,dim],[dim,dim],depth=depth,width=width)
	analysis(f"depth={depth}, width={width}",mult,dim,N=2**14,epochs=2**12,training_σ=10)

plt.xscale('log')
plt.ylabel("log(MSE)")
plt.xlabel("epoch")
plt.title("scalar multiplication by width")
plt.legend()
plt.savefig("out/matmul_scalar_by_width.png")






dim=1
depth=5
width=8
mult = MatMul_0([dim,dim],[dim,dim],depth=depth,width=width)
analysis(f"depth={depth}, width={width}",mult,dim,N=2**14,epochs=2**12,training_σ=10)

from PIL import Image

def error_mult(name,mult, x_range=[-100,100],y_range=[-100,100]):
	# x_range=[-60,60]
	# y_range=[-60,60]
	resolution = 2**10
	x_span = torch.linspace(x_range[0],x_range[1],resolution,device=device)
	y_span = torch.linspace(y_range[0],y_range[1],resolution,device=device)
	#
	x,y = torch.meshgrid(x_span,y_span)
	x = x.reshape(-1,1,1) # .reshape(2**11,2**11)
	y = y.reshape(-1,1,1) # .reshape(2**11,2**11)
	sgn       = ((mult(x,y)*(x*y) > 0)*2 - 1  ).reshape(resolution,resolution).detach().cpu().numpy()
	err_add   = ((mult(x,y) - (x*y))          ).reshape(resolution,resolution).detach().cpu().numpy()
	err_ratio = ((mult(x,y) / (x*y))-1        ).reshape(resolution,resolution).detach().cpu().numpy()
	# xy = torch.log(mult(x,y)) - torch.log(x*y)
	# sgn       = sgn      .reshape(resolution,resolution).detach().cpu().numpy()
	# err_add   = err_add  .reshape(resolution,resolution).detach().cpu().numpy()
	# err_ratio = err_ratio.reshape(resolution,resolution).detach().cpu().numpy()
	# rgb = np.stack((err_ratio*(err_ratio>0),np.sin(3.115*3*err_ratio)>0.9,-err_ratio*(err_ratio<0)),-1)
	rgb = np.stack((err_ratio*(err_ratio>0),err_ratio*0,-err_ratio*(err_ratio<0)),-1)
	Image.fromarray((rgb.clip(0,1)*255).astype(np.uint8)).save(f"out/{name}_error_ratio.png")
	rgb = np.stack((err_add  *(err_add  >0),err_add  *0,-err_add  *(err_add  <0)),-1)
	Image.fromarray((rgb.clip(0,1)*255).astype(np.uint8)).save(f"out/{name}_error_addition.png")
	rgb = np.stack((sgn      *(sgn      >0),sgn      *0,-sgn      *(sgn      <0)),-1)
	Image.fromarray((rgb.clip(0,1)*255).astype(np.uint8)).save(f"out/{name}_error_sgn.png")
	# plt.imshow(rgb.transpose(1,0,2),extent=(x_range[0],x_range[1],y_range[0],y_range[1]),origin="lower")
	# return Image.fromarray((rgb.clip(0,1)*255).astype(np.uint8))
	# plt.contour(np.abs( np.exp(z*np.cos(θ)*np.cos(θ/2)) ), alpha=0.5)

name = f"matmul_depth={depth}_width={width}"
error_mult(name,mult)


resolution = 2**10
for N in [0.01,0.1,1,10,100,1000]:
	x = torch.linspace(-70,70,resolution,device=device).reshape(-1,1,1)
	y = N*torch.ones(x.shape,device=device)
	# plt.clf()
	plt.close("all");plt.figure(figsize=[3.2,2.4])
	plt.plot(x.reshape(-1).cpu(),(x*y).reshape(-1).detach().cpu(),label=f"x⋅{N}")
	plt.plot(x.reshape(-1).cpu(),mult(x,y).reshape(-1).detach().cpu(),label=f"model(x,{N})")
	plt.plot(x.reshape(-1).cpu(),mult(y,x).reshape(-1).detach().cpu(),label=f"model({N},x)")
	plt.legend()
	plt.savefig(f"out/{name}_perf_{N}_x.png")

for N in [0.01,0.1,1,10,100,1000]:
	x = torch.linspace(-70,70,resolution,device=device).reshape(-1,1,1)
	y = N*torch.ones(x.shape,device=device)
	# plt.clf()
	plt.close("all");plt.figure(figsize=[3.2,2.4])
	plt.plot(x.reshape(-1).cpu(),N*torch.ones(x.shape).reshape(-1).detach().cpu(),label=f"{N}")
	plt.plot(x.reshape(-1).cpu(),mult(x,N/x).reshape(-1).detach().cpu(),label=f"model(x,{N}/x)")
	plt.plot(x.reshape(-1).cpu(),mult(N/x,x).reshape(-1).detach().cpu(),label=f"model({N}/x,x)")
	plt.legend()
	plt.ylim([N*1.5+5,N*0.5-5])
	plt.savefig(f"out/{name}_perf_{N}_x_x.png")

for N in [0.01,0.1,1,10,100,1000]:
	x = torch.linspace(-70,70,resolution,device=device).reshape(-1,1,1)
	# plt.clf()
	# plt.close("all");plt.figure(figsize=[3.2,2.4])
	# plt.plot(x.reshape(-1).cpu(),-(x*x).reshape(-1).detach().cpu(),label=f"-x⋅x")
	# plt.plot(x.reshape(-1).cpu(),+(x*x).reshape(-1).detach().cpu(),label=f"x⋅x")
	# plt.plot(x.reshape(-1).cpu(),mult(x,-x).reshape(-1).detach().cpu(),label=f"model(x,-x)")
	# plt.plot(x.reshape(-1).cpu(),mult(-x,x).reshape(-1).detach().cpu(),label=f"model(-x,x)")
	# plt.savefig(f"out/{name}_perf_x2.png")
	plt.close("all");plt.figure(figsize=[3.2,2.4])
	plt.plot(x.reshape(-1).cpu(),N*(x*x).reshape(-1).detach().cpu(),label=f"N⋅x⋅x")
	plt.plot(x.reshape(-1).cpu(),mult(x,N*x).reshape(-1).detach().cpu(),label=f"model(x,{N}⋅x)")
	plt.plot(x.reshape(-1).cpu(),mult(N*x,x).reshape(-1).detach().cpu(),label=f"model({N}⋅x,x)")
	plt.legend()
	plt.savefig(f"out/{name}_perf_{N}_x2.png")


# plt.plot(x.reshape(-1).cpu(),(x*x).reshape(-1).detach().cpu())
# plt.plot(x.reshape(-1).cpu(),mult(x,x).reshape(-1).detach().cpu())

# plt.plot(x.reshape(-1).cpu(),(-x*x).reshape(-1).detach().cpu())
# plt.plot(x.reshape(-1).cpu(),mult(-x,x).reshape(-1).detach().cpu())
# plt.plot(x.reshape(-1).cpu(),mult(x,-x).reshape(-1).detach().cpu())






# plt.clf()
# show_c(lambda z: z, x_range=[-100,100],y_range=[-100,100],r_freq=1,θ_freq=16)
# show_c(fun, ,r_freq=16,θ_freq=30)













dim=2
depth=5
width=8
name2d = f"matmul_depth={depth}_width={width}_dim={dim}"
mult2d = MatMul_0([dim,dim],[dim,dim],depth=depth,width=width)
plt.close("all");plt.figure()
analysis(name2d,mult2d,dim,N=2**14,epochs=2**14,lr0=-1,lr1=-2)
plt.xscale('log')
plt.ylabel(f"log(MSE)")
plt.xlabel(f"epoch")
plt.title(f"mat2 multiplication")
plt.legend()
plt.savefig(f"out/{name2d}.png")











A = torch.zeros(2,2,device=device).fill_diagonal_(1).unsqueeze(0)
B = torch.randn(2,2,device=device).unsqueeze(0)

A = torch.randn(2,2,device=device).unsqueeze(0)
B = torch.randn(2,2,device=device).unsqueeze(0)

nn.MSELoss()(mult2d(A,B),A@B).item()
torch.mean(torch.abs(mult2d(A,B) - A@B)**2)










































def trainer(name,mult,dim_in,dim_out,N=2**14,epochs=2**12,lr0=-1,lr1=-2, plot=True, training_σ=1):
	mult = mult.to(device)
	lrs = 10**(torch.linspace(lr0,lr1,epochs))
	optim = torch.optim.Adam(mult.parameters(),lr=1e-2,eps=1e-12)
	hist = []
	for epoch in range(epochs):
		mult.train()
		lr = (lrs[epoch]).item()
		for g in optim.param_groups: g['lr'] = lr
		### many matrices
		A = torch.randn(N,dim_in,dim_out,device=device)*training_σ # 10
		X = torch.randn(N,dim_in,1,device=device)*training_σ # 10
		### matrix multiplication
		y = A@X # ground truth
		Y = mult(A,X)
		### training
		loss = nn.MSELoss()(y,Y)
		optim.zero_grad()
		loss.backward()
		optim.step()
		hist.append(loss.item())
	if plot: plt.scatter(range(len(hist)),np.log(hist),s=.05,label=name)
	mult.hist = hist
	return mult

dim_in=2;dim_out=2
# dim_in=3;dim_out=3
# dim_in=4;dim_out=4
depth=5
width=8
name2d = f"matmul_depth={depth}_width={width}_dim_in={dim_in}_dim_out={dim_out}"
mult2d = MatMul_0([dim_in,dim_out],[dim_out,1],depth=depth,width=width)
plt.close("all");plt.figure()
trainer(name2d,mult2d,dim_in,dim_out,N=2**14,epochs=2**14,lr0=-1,lr1=-2)
plt.xscale('log')
plt.ylabel(f"log(MSE)")
plt.xlabel(f"epoch")
plt.title(f"mat2⋅vec2")
plt.legend()
plt.grid()
plt.savefig(f"out/{name2d}.png")


# A = torch.zeros(dim_out,dim_in,device=device).fill_diagonal_(1).unsqueeze(0)
# B = torch.randn(dim_in,1,device=device).unsqueeze(0)

# A = torch.randn(2,2,device=device).unsqueeze(0)
# B = torch.randn(2,1,device=device).unsqueeze(0)

# nn.MSELoss()(mult2d(A,B),A@B).item()
# torch.mean(torch.abs(mult2d(A,B) - A@B)**2)

# nn.MSELoss()(mult2d(A,B),A@B).item()

# decomposed = mult2d(A,B*torch.tensor([[[1],[0]]],device=device)) + mult2d(A,B*torch.tensor([[[0],[1]]],device=device))
# direct     = mult2d(A,B)



N=2**16
A = torch.randn(N,dim_out,dim_in,device=device)
B = torch.randn(N,dim_in,1,device=device)

truth = A@B
found = mult2d(A,B)

error = (truth-found)**2
error = torch.log(torch.sqrt(error.sum(dim=2).sum(dim=1)))
error = error.cpu().detach()

plt.close("all");plt.figure(figsize=[3.2,2.4])
plt.hist(error,256,density=True)
# plt.xscale("log")
plt.xlabel("log(error)")
plt.ylabel("distribution")
plt.tight_layout()
plt.savefig(f"out/{name2d}_error.png")
# plt.show()


# nn.MSELoss()(mult2d(A,B),A@B).item()




# dot = (truth*truth).sum(2).sum(1) / ((truth**2).sum(2).sum(1)) # debug : must be 1
dot = (truth*found).sum(2).sum(1) / ((truth**2).sum(2).sum(1))
dot = dot.cpu().detach()

plt.close("all");plt.figure(figsize=[3.2,2.4])
plt.hist(dot,512,density=True)
plt.plot([1,1],[0,1])
# plt.plot([0,0],[0,1])
plt.yscale("log")
plt.grid()
# plt.show()

plt.tight_layout()
plt.savefig(f"out/{name2d}_dot.png")

plt.xlim([-10,10])
plt.tight_layout()
plt.savefig(f"out/{name2d}_dot_zoom.png")










B1 = torch.randn(N,dim_in,1,device=device)
B2 = torch.randn(N,dim_in,1,device=device)

direct = mult2d(A,B1+B2)
indirect = mult2d(A,B1)+mult2d(A,B2)
truth = A@(B1+B2)

dir_indir = torch.log(torch.sqrt(((direct-indirect)**2).sum(2).sum(1)))
truth_dir = torch.log(torch.sqrt(((truth-direct)**2).sum(2).sum(1)))
truth_indir = torch.log(torch.sqrt(((truth-indirect)**2).sum(2).sum(1)))

# plt.close("all");plt.figure(figsize=[3.2,2.4])
plt.close("all");plt.figure()
plt.hist(dir_indir  .cpu().detach(),256,density=True,label="log([model(A,B1+B2)] - [model(A,B1)+model(A,B2)])",alpha=0.5)
plt.hist(truth_dir  .cpu().detach(),256,density=True,label="log([A⋅(B1+B2)] - [model(A,B1+B2)])",alpha=0.5)
plt.hist(truth_indir.cpu().detach(),256,density=True,label="log([A⋅(B1+B2)] - [model(A,B1)+model(A,B2)])",alpha=0.5)
plt.legend(loc="lower left")
# plt.xlabel("log([model(A,B1+B2)] - [model(A,B1)+model(A,B2)])")
# plt.plot([1,1],[0,1])
# # plt.plot([0,0],[0,1])
plt.yscale("log")
# plt.grid()
# plt.show()
plt.tight_layout()
plt.savefig(f"out/{name2d}_propt.png")





























from complex import C_Log, Exp, ToReal, ToComplex
from linear import *

# plt.clf()
# dim=1
# mult = MatMul_0([dim,dim],[dim,dim],depth=0,width=4);analysis(mult,dim,N=2**12,epochs=2**11)
# mult = MatMul_0([dim,dim],[dim,dim],depth=1,width=4);analysis(mult,dim,N=2**12,epochs=2**11)
# mult = MatMul_0([dim,dim],[dim,dim],depth=2,width=4);analysis(mult,dim,N=2**12,epochs=2**11)
# mult = MatMul_0([dim,dim],[dim,dim],depth=3,width=4);analysis(mult,dim,N=2**12,epochs=2**11)
# mult = MatMul_0([dim,dim],[dim,dim],depth=4,width=4);analysis(mult,dim,N=2**12,epochs=2**11)
# mult = MatMul_0([dim,dim],[dim,dim],depth=8,width=4);analysis(mult,dim,N=2**12,epochs=2**11)



class Mul_Log(nn.Module):
	# def __init__(self):
	# 	super().__init__()
	# 	self.model = nn.Sequential(
	# 		ToComplex(), C_Log(),
	# 		C_Linear_Binary(2,1),
	# 		Exp(), ToReal(),
	# 	)
	def __init__(self,dim_left,dim_right,width=1,log_mixer_mode="binary",out_mixer_mode="2softmax"):
		super().__init__()
		if not dim_left[1]==dim_right[0]: print("ERROR INCOMPATIBILITY DETECTED")
		self.dim_left = dim_left
		self.dim_right = dim_right
		self.flatten = nn.Flatten()
		dim_in  = dim_left[0]*dim_left[1] + dim_right[0]*dim_right[1]
		dim_mid = width*dim_in
		dim_out = dim_left[0]*dim_right[1]
		if log_mixer_mode=="linear"  : log_mixer = C_Linear(dim_in,dim_mid)
		if log_mixer_mode=="binary"  : log_mixer = C_Linear_Binary(dim_in,dim_mid)
		if log_mixer_mode=="sigmoid" : log_mixer = C_Linear_Binary_Sigmoid(dim_in,dim_mid)
		if log_mixer_mode=="linear"  : log_mixer = C_Linear(dim_in,dim_mid,bias=False)
		if log_mixer_mode=="2softmax": log_mixer = C_Linear_2_Softmax(dim_in,dim_mid)
		# if log_mixer_mode=="ternary" : log_mixer = C_Linear_Ternary(dim_in,dim_mid)
		if log_mixer_mode=="3sigmoid": log_mixer = C_Linear_Ternary_Sigmoid(dim_in,dim_mid)
		###
		if out_mixer_mode=="linear"  : out_mixer = C_Linear(dim_mid,dim_out)
		if out_mixer_mode=="binary"  : out_mixer = C_Linear_Binary(dim_mid,dim_out)
		if out_mixer_mode=="sigmoid" : out_mixer = C_Linear_Binary_Sigmoid(dim_mid,dim_out)
		if out_mixer_mode=="linear"  : out_mixer = C_Linear(dim_mid,dim_out,bias=False)
		if out_mixer_mode=="2softmax": out_mixer = C_Linear_2_Softmax(dim_mid,dim_out)
		# if out_mixer_mode=="ternary" : out_mixer = C_Linear_Ternary(dim_mid,dim_out)
		if out_mixer_mode=="3sigmoid": out_mixer = C_Linear_Ternary_Sigmoid(dim_mid,dim_out)
		self.model = nn.Sequential(
			ToComplex(), C_Log(),
			log_mixer,
			Exp(), ToReal(),
			# C_Linear_Binary(dim_mid,dim_out),
			# nn.Linear(dim_mid,dim_out), # TODO Linear_Integer
			# C_Linear_2_Softmax(dim_mid,dim_out),
			# C_Linear_Ternary(dim_mid,dim_out),
			# C_Linear_Ternary_Sigmoid(dim_mid,dim_out)
			out_mixer
		)
	def forward(self,A,B):
		# return self.model(torch.hstack((A,B)))
		AB = torch.hstack((self.flatten(A),self.flatten(B)))
		return self.model(AB).view(-1,self.dim_left[0],self.dim_right[1])


dim=2
dim_A = [dim,dim]
dim_B = [dim,dim]
log_mixer_mode="linear"
out_mixer_mode="linear"
width=2
mult = Mul_Log([dim,dim],[dim,dim],width=width,log_mixer_mode=log_mixer_mode,out_mixer_mode=out_mixer_mode)
name = f"dim_A={dim_A}-dim_A={dim_B}-width={width}-log_mixer={log_mixer_mode}-out_mixer{out_mixer_mode}"
analysis(name,mult,dim,N=2**14,epochs=2**12,plot=False)
plt.clf()
plt.scatter(range(len(mult.hist)),np.log(mult.hist),s=.05,label=f"{log_mixer_mode} → {out_mixer_mode}")
plt.show()






dim=3
width=dim
dim_A = [dim,dim]
dim_B = [dim,dim]
mults={}
plt.clf()
plt.ioff()
for m in ["linear", "binary","sigmoid","linear","2softmax","3sigmoid",]:
	mults[m]={}
	for n in ["linear", "binary","sigmoid","linear","2softmax","3sigmoid",]:
		mult = Mul_Log([dim,dim],[dim,dim],width=width,log_mixer_mode=m,out_mixer_mode=n)
		name = f"dim_A={dim_A}-dim_A={dim_B}-width={width}-log_mixer={m}-out_mixer{n}"
		analysis(name,mult,dim,N=2**14,epochs=2**12,plot=False)
		mults[m][n] = mult
		# mults[m][n] = {"hist": mult.hist}

for m in ["linear", "binary","sigmoid","linear","2softmax","3sigmoid",]:
	plt.clf()
	# plt.figure()
	for n in ["linear", "binary","sigmoid","linear","2softmax","3sigmoid",]:
		mult = mults[m][n]
		plt.scatter(range(len(mult.hist)),np.log(mult.hist),s=.05,label=f"{m} → {n}")
	plt.legend()
	os.makedirs(f"out/",exist_ok=True)
	plt.tight_layout()
	plt.savefig(f"out/mult-dim={dim}-width={width}-log_mixer={m}.png")



for n in ["linear", "binary","sigmoid","linear","2softmax","3sigmoid",]:
	plt.clf()
	for m in ["linear", "binary","sigmoid","linear","2softmax","3sigmoid",]:
		mult=mults[m][n]
		plt.scatter(range(len(mult.hist)),np.log(mult.hist),s=.05,label=f"{m} → {n}")
	plt.legend()
	plt.tight_layout()
	plt.savefig(f"out/mult-dim={dim}-width={width}-out_mixer={n}.png")


# mult = Mul_Log([dim,dim],[dim,dim],width=2,log_mixer_mode="sigmoid") ;analysis(mult,dim,N=2**12,epochs=2**11)
# mult = Mul_Log([dim,dim],[dim,dim],width=2,log_mixer_mode="2softmax");analysis(mult,dim,N=2**12,epochs=2**11)
# mult = Mul_Log([dim,dim],[dim,dim],width=2,log_mixer_mode="binary")  ;analysis(mult,dim,N=2**12,epochs=2**11)
# mult = Mul_Log([dim,dim],[dim,dim],width=2,log_mixer_mode="linear")  ;analysis(mult,dim,N=2**12,epochs=2**11)

plt.clf()
viz_error(mult,dim,N=2**14)

# mult = Mul_Log([dim,dim],[dim,dim],width=1);analysis(mult,dim,N=2**12,epochs=2**11)
# mult = Mul_Log([dim,dim],[dim,dim],width=2);analysis(mult,dim,N=2**12,epochs=2**11)
# mult = Mul_Log([dim,dim],[dim,dim],width=3);analysis(mult,dim,N=2**12,epochs=2**11)



# dim=2


# analysis(MatMul_0(dim   ),dim,N=2**15,epochs=2**12)

# analysis(MatMul_0b(dim,1 ,0),dim,N=2**15,epochs=2**14)
# analysis(MatMul_0b(dim,2 ,0),dim,N=2**15,epochs=2**14)
# analysis(MatMul_0b(dim,3 ,0),dim,N=2**15,epochs=2**14)
# analysis(MatMul_0b(dim,5 ,0),dim,N=2**15,epochs=2**12)
# analysis(MatMul_0b(dim,10,0),dim,N=2**15,epochs=2**12)
# analysis(MatMul_0b(dim,20,0),dim,N=2**15,epochs=2**12)

# analysis(MatMul_0b(dim,1 ),dim,N=2**15,epochs=2**14)
# analysis(MatMul_0b(dim,2 ),dim,N=2**15,epochs=2**14)
# analysis(MatMul_0b(dim,3 ),dim,N=2**15,epochs=2**14)
# analysis(MatMul_0b(dim,5 ),dim,N=2**15,epochs=2**14)
# analysis(MatMul_0b(dim,10),dim,N=2**15,epochs=2**14)
# analysis(MatMul_0b(dim,20),dim,N=2**15,epochs=2**14)
